from google.appengine.ext import ndb
import logging

"""
    db_actor handles all db operations for an actor
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_actor',
    'db_actor_list',
]


class Actor(ndb.Model):
    """
       NDB data model for an actor
    """
    id = ndb.StringProperty(required=True)
    creator = ndb.StringProperty()
    passphrase = ndb.StringProperty()


class db_actor():
    """
        db_actor does all the db operations for actor objects

    """

    def get(self,  actorId=None):
        """ Retrieves the actor from the database """
        if not actorId:
            return None
        self.handle = Actor.query(Actor.id == actorId).get(use_cache=False)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "creator": t.creator,
                "passphrase": t.passphrase,
            }
        else:
            return None

    def getByCreator(self, creator=None):
        """ Retrieves the actor from db based on creator field

            Returns None if none was found. If one is found, that one is
            loaded in the object. If more, the last one found is loaded.
        """
        if not creator:
            return None
        self.handle = Actor.query(Actor.creator == creator).fetch(use_cache=False)
        if not self.handle or len(self.handle) == 0:
            return None
        ret = []
        if len(self.handle) == 1:
            ret.append(self.get(actorId=self.handle[0].id))
            return ret
        logging.warn("Found multiple actors with creator(" + creator + "):")
        for c in self.handle:
            logging.warn("    id (" + c.id + ")")
            ret.append(self.get(actorId=c.id))
        return ret

    def modify(self, creator=None, passphrase=None):
        """ Modify an actor """
        if not self.handle:
            logging.debug("Attempted modification of db_actor without db handle")
            return False
        if creator and len(creator) > 0:
            self.handle.creator = creator
        if passphrase and len(passphrase) > 0:
            self.handle.passphrase = passphrase
        self.handle.put(use_cache=False)
        return True

    def create(self, actorId=None, creator=None,
               passphrase=None):
        """ Create a new actor """
        if not actorId:
            return False
        if not creator:
            creator = ''
        if not passphrase:
            passphrase = ''
        self.handle = Actor(id=actorId,
                            creator=creator,
                            passphrase=passphrase)
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the actor in the database """
        if not self.handle:
            logging.debug("Attempted delete of db_actor without db handle")
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None


class db_actor_list():
    """
        db_actor_list does all the db operations for list of actor objects
    """

    def fetch(self):
        """ Retrieves the actors in the database """
        self.handle = Actor.query().fetch(use_cache=False)
        if self.handle:
            ret = []
            for t in self.handle:
                ret.append({
                    "id": t.id,
                    "creator": t.creator,
                    })
            return ret
        else:
            return False

    def __init__(self):
        self.handle = None



from actingweb.handlers import base_handler

class callback_oauth_handler(base_handler.base_handler):

    def get(self):
        if not self.request.get('code'):
            self.response.set_status(400, "Bad request. No code.")
            return
        code = self.request.get('code')
        id = self.request.get('state')
        self.response.set_redirect(self.config.root + str(id) + '/oauth?code=' + str(code))




import urllib
import logging
import json

# This function code is from latest urlfetch. For some reason the
# Appengine version of urlfetch does not include links()


def PaginationLinks(self):
    """Links parsed from HTTP Link header"""
    ret = []
    if 'link' in self.headers:
        linkheader = self.headers['link']
    else:
        return ret
    for i in linkheader.split(','):
        try:
            url, params = i.split(';', 1)
        except ValueError:
            url, params = i, ''
        link = {}
        link['url'] = url.strip('''<> '"''')
        for param in params.split(';'):
            try:
                k, v = param.split('=')
            except ValueError:
                break
            link[k.strip(''' '"''')] = v.strip(''' '"''')
        ret.append(link)
    return ret


class oauth():

    def __init__(self, token=None, config=None):
        self.config = config
        self.token = token
        self.first = None
        self.next = None
        self.prev = None
        self.last_response_code = 0
        self.last_response_message = ""

    def enabled(self):
        if len(self.config.oauth['client_id']) == 0:
            return False
        else:
            return True

    def setToken(self, token):
        if token:
            self.token = token

    def postRequest(self, url, params=None, urlencode=False):
        if params:
            if urlencode:
                data = urllib.urlencode(params)
                logging.info('Oauth POST request with urlencoded payload: ' + url + ' ' + data)
            else:
                data = json.dumps(params)
                logging.info('Oauth POST request with JSON payload: ' + url + ' ' + data)
        else:
            data = None
            logging.info('Oauth POST request: ' + url)
        if urlencode:
            if self.token:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           }
        else:
            if self.token:
                headers = {'Content-Type': 'application/json',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/json'}
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    payload=data,
                    method=self.config.module["urlfetch"].POST,
                    headers=headers)
            else:
                response = self.config.module["urlfetch"].post(url=url, data=data, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except:
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warn("Oauth POST failed with exception")
            return None
        if response.status_code == 204:
            return {}
        if response.status_code != 200 and response.status_code != 201:
            logging.info('Error when sending POST request: ' +
                         str(response.status_code) + response.content)
            return None
        logging.debug('Oauth POST response JSON:' + response.content)
        return json.loads(response.content)

    def putRequest(self, url, params=None, urlencode=False):
        if params:
            if urlencode:
                data = urllib.urlencode(params)
                logging.info('Oauth PUT request with urlencoded payload: ' + url + ' ' + data)
            else:
                data = json.dumps(params)
                logging.info('Oauth PUT request with JSON payload: ' + url + ' ' + data)
        else:
            data = None
            logging.info('Oauth PUT request: ' + url)
        if urlencode:
            if self.token:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/x-www-form-urlencoded',
                           }
        else:
            if self.token:
                headers = {'Content-Type': 'application/json',
                           'Authorization': 'Bearer ' + self.token,
                           }
            else:
                headers = {'Content-Type': 'application/json'}
        try:
            if self.config.env == 'appengine':
                self.config.module["urlfetch"].set_default_fetch_deadline(20)
                response = self.config.module["urlfetch"](
                    url=url,
                    payload=data,
                    method=self.config.module["urlfetch"].PUT,
                    headers=headers)
            else:
                response = self.config.module["urlfetch"].post(url=url, data=data, headers=headers)
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except:
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warn("Oauth PUT failed with exception")
            return None
        if response.status_code == 204:
            return {}
        if response.status_code != 200 and response.status_code != 201:
            logging.info('Error when sending PUT request: ' +
                         str(response.status_code) + response.content)
            return None
        logging.debug('Oauth PUT response JSON:' + response.content)
        return json.loads(response.content)

    def getRequest(self, url, params=None):
        if not self.token:
            logging.debug("No token set in getRequest()")
            return None
        if params:
            url = url + '?' + urllib.urlencode(params)
        logging.info('Oauth GET request: ' + url)
        try:
            if self.config.env == 'appengine':
                urlfetch.self.config.module["urlfetch"].set_default_fetch_deadline(60)
                response = self.config.module["urlfetch"].fetch(
                    url=url,
                    method=self.config.module["urlfetch"].GET,
                    headers={
                        'Authorization': 'Bearer ' + self.token,
                    }
                    )
            else:
                response = self.config.module["urlfetch"].get(
                    url=url,
                    headers={
                        'Authorization': 'Bearer ' + self.token,
                    }
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except:
            self.last_response_code = 0
            self.last_response_message = 'No response'
            logging.warn("Oauth GET failed with exception")
            return None
        if response.status_code < 200 or response.status_code > 299:
            logging.info('Error when sending GET request to Oauth: ' +
                         str(response.status_code) + response.content)
            return None
        links = PaginationLinks(response)
        self.next = None
        self.first = None
        self.prev = None
        for link in links:
            logging.debug('Links:' + link['rel'] + ':' + link['url'])
            if link['rel'] == 'next':
                self.next = link['url']
            elif link['rel'] == 'first':
                self.first = link['url']
            elif link['rel'] == 'prev':
                self.prev = link['url']
        return json.loads(response.content)

    def deleteRequest(self, url):
        if not self.token:
            return None
        logging.info('Oauth DELETE request: ' + url)
        try:
            if self.config.env == 'appengine':
                response = self.config.module["urlfetch"].fetch(url=url,
                                          method=self.config.module["urlfetch"].DELETE,
                                          headers={'Authorization': 'Bearer ' + self.token}
                                          )
            else:
                response = self.config.module["urlfetch"].delete(
                    url=url,
                    headers={'Authorization': 'Bearer ' + self.token}
                    )
            self.last_response_code = response.status_code
            self.last_response_message = response.content
        except:
            logging.warn("Spark DELETE failed.")
            self.last_response_code = 0
            self.last_response_message = 'No response'
            return None
        if response.status_code < 200 or response.status_code > 299:
            logging.info('Error when sending DELETE request to Oauth: ' +
                         str(response.status_code) + response.content)
            return None
        if response.status_code == 204:
            return {}
        try:
            ret = json.loads(response.content)
        except:
            return {}
        return ret

    def oauthRedirectURI(self, state=''):
        params = {
            'response_type': self.config.oauth['response_type'],
            'client_id': self.config.oauth['client_id'],
            'redirect_uri': self.config.oauth['redirect_uri'],
            'scope': self.config.oauth['scope'],
            'state': state,
        }
        uri = self.config.oauth['auth_uri'] + "?" + urllib.urlencode(params)
        logging.info('OAuth redirect with url: ' + uri + ' and state:' + state)
        return uri

    def oauthRequestToken(self, code=None):
        if not code:
            return None
        params = {
            'grant_type': self.config.oauth['grant_type'],
            'client_id': self.config.oauth['client_id'],
            'client_secret': self.config.oauth['client_secret'],
            'code': code,
            'redirect_uri': self.config.oauth['redirect_uri'],
        }
        self.token = None
        result = self.postRequest(url=self.config.oauth[
                                  'token_uri'], params=params, urlencode=True)
        if result and 'access_token' in result:
            self.token = result['access_token']
        return result

    def oauthRefreshToken(self, refresh_token):
        if not refresh_token:
            return None
        params = {
            'grant_type': self.config.oauth['refresh_type'],
            'client_id': self.config.oauth['client_id'],
            'client_secret': self.config.oauth['client_secret'],
            'refresh_token': refresh_token,
        }
        self.token = None
        result = self.postRequest(url=self.config.oauth[
                                  'token_uri'], params=params, urlencode=True)
        if not result:
            self.token = None
            return None
        self.token = result['access_token']
        return result



from google.appengine.ext import ndb
import logging

"""
    db_trust handles all db operations for a trust
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_trust',
    'db_trust_list',
]


class Trust(ndb.Model):
    """ Data model for a trust relationship """
    id = ndb.StringProperty(required=True)
    peerid = ndb.StringProperty(required=True)
    baseuri = ndb.StringProperty(required=True)
    type = ndb.StringProperty(required=True)
    relationship = ndb.StringProperty(required=True)
    secret = ndb.StringProperty(required=True)
    desc = ndb.TextProperty()
    approved = ndb.BooleanProperty()
    peer_approved = ndb.BooleanProperty()
    verified = ndb.BooleanProperty()
    verificationToken = ndb.StringProperty()


class db_trust():
    """
        db_trust does all the db operations for trust objects

        The  actorId must always be set.
    """

    def get(self,  actorId=None, peerid=None, token=None):
        """ Retrieves the property from the database

            Either peerid or token must be set.
            If peerid is set, token will be ignored.
        """
        if not actorId:
            return None
        if not self.handle and peerid:
            self.handle = Trust.query(Trust.id == actorId,
                                      Trust.peerid == peerid).get(use_cache=False)
        elif not self.handle and token:
            self.handle = Trust.query(Trust.id == actorId,
                                      Trust.secret == token).get(use_cache=False)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "peerid": t.peerid,
                "baseuri": t.baseuri,
                "type": t.type,
                "relationship": t.relationship,
                "secret": t.secret,
                "desc": t.desc,
                "approved": t.approved,
                "peer_approved": t.peer_approved,
                "verified": t.verified,
                "verificationToken": t.verificationToken,
            }
        else:
            return None

    def modify(self, baseuri=None,
               secret=None,
               desc=None,
               approved=None,
               verified=None,
               verificationToken=None,
               peer_approved=None):
        """ Modify a trust 
        
            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of db_trust without db handle")
            return False
        if baseuri and len(baseuri) > 0:
            self.handle.baseuri = baseuri
        if secret and len(secret) > 0:
            self.handle.secret = secret
        if desc and len(desc) > 0:
            self.handle.desc = desc
        if approved is not None:
            self.handle.approved = approved
        if verified is not None:
            self.handle.verified = verified
        if verificationToken and len(verificationToken) > 0:
            self.handle.verificationToken = verificationToken
        if peer_approved is not None:
            self.handle.peer_approved = peer_approved
        self.handle.put(use_cache=False)
        return True

    def create(self, actorId=None, peerid=None,
               baseuri='', type='', relationship='',
               secret='', approved='', verified=False,
               peer_approved=False, verificationToken='',
               desc=''):
        """ Create a new trust """
        if not actorId or not peerid:
            return False
        self.handle = Trust(id=actorId,
                            peerid=peerid,
                            baseuri=baseuri,
                            type=type,
                            relationship=relationship,
                            secret=secret,
                            approved=approved,
                            verified=verified,
                            peer_approved=peer_approved,
                            verificationToken=verificationToken,
                            desc=desc)
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the property in the database """
        if not self.handle:
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def isTokenInDB(self, actorId=None, token=None):
        """ Returns True if token is found in db """
        if not actorId or len(actorId) == 0:
            return False
        if not token or len(token) == 0:
            return False
        res = Trust.query(Trust.id == actorId, 
                          Trust.secret == token).get(use_cache=False)
        if res:
            return True
        return False

    def __init__(self):
        self.handle = None


class db_trust_list():
    """
        db_trust_list does all the db operations for list of trust objects

        The  actorId must always be set. 
    """

    def fetch(self,  actorId):
        """ Retrieves the trusts of an actorId from the database as an array"""
        if not actorId:
            return None
        self.handle = Trust.query(Trust.id == actorId).fetch(use_cache=False)
        self.trusts = []
        if self.handle:
            for t in self.handle:
                self.trusts.append(
                {
                    "id": t.id,
                    "peerid": t.peerid,
                    "baseuri": t.baseuri,
                    "type": t.type,
                    "relationship": t.relationship,
                    "secret": t.secret,
                    "desc": t.desc,
                    "approved": t.approved,
                    "peer_approved": t.peer_approved,
                    "verified": t.verified,
                    "verificationToken": t.verificationToken,
                })
            return self.trusts
        else:
            return []

    def delete(self):
        """ Deletes all the properties in the database """
        if not self.handle:
            return False
        for p in self.handle:
            p.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None



class attributes():
    """
        attribute is the main entity keeping an attribute.

        It needs to be initalized at object creation time.

    """

    def get_bucket(self):
        """ Retrieves the attribute bucket from the database """
        if not self.data or len(self.data) == 0:
            self.data = self.dbprop.get_bucket(actorId=self.actorId, bucket=self.bucket)
        return self.data

    def get_attr(self, name=None):
        """ Retrieves a single attribute """
        if not name:
            return None
        if not name in self.data:
            self.data[name] = self.dbprop.get_attr(actorId=self.actorId, bucket=self.bucket, name=name)
        return self.data[name]

    def set_attr(self, name=None, data=None, timestamp=None):
        """ Sets new data for this attribute """
        if not self.actorId or not self.bucket:
            return False
        if not name in data:
            self.data[name] = {}
        self.data[name]["data"] = data
        self.data[name]["timestamp"] = timestamp
        return self.dbprop.set_attr(
            actorId=self.actorId,
            bucket=self.bucket,
            name=name,
            data=data,
            timestamp=timestamp
        )

    def delete_attr(self, name=None):
        if not name:
            return False
        del self.data[name]
        return self.dbprop.delete_attr(actorId=self.actorId, bucket=self.bucket, name=name)

    def delete_bucket(self):
        """ Deletes the attribute bucket in the database """
        if not self.dbprop:
            return False
        if self.dbprop.delete_bucket(actorId=self.actorId, bucket=self.bucket):
            self.dbprop = self.config.db_attribute.db_attribute()
            self.data = {}
            return True
        else:
            return False

    def __init__(self,  actorId=None, bucket=None, config=None):
        """ A attribute must be initialised with actorId and bucket
        """
        self.config = config
        self.dbprop = self.config.db_attribute.db_attribute()
        self.bucket = bucket
        self.actorId = actorId
        self.data = {}
        if actorId and bucket and len(bucket) > 0 and config:
            self.get_bucket()


class buckets():
    """ Handles all attribute buckets of a specific actor_id

        Access the attributes
        in .props as a dictionary
    """

    def fetch(self):
        if not self.actorId:
            return False
        return self.list.fetch(actorId=self.actorId)

    def fetch_timestamps(self):
        if not self.actorId:
            return False
        return self.list.fetch_timestamps(actorId=self.actorId)

    def delete(self):
        if not self.list:
            return False
        self.list.delete(actorId=self.actorId)
        self.list = self.config.db_attribute.db_attribute_bucket_list()
        return True

    def __init__(self,  actorId=None, config=None):
        """ attributes must always be initialised with an actorId """
        self.config = config
        if not actorId:
            self.list = None
            return
        self.list = self.config.db_attribute.db_attribute_bucket_list()
        self.actorId = actorId





import logging

class trust():

    def get(self):
        """ Retrieve a trust relationship with either peerid or token """
        if self.trust and len(self.trust) > 0:
            return self.trust
        if not self.peerid and self.token:
            self.trust = self.handle.get(actorId=self.actorId,
                                         token=self.token)
        elif self.peerid and not self.token:
            self.trust = self.handle.get(actorId=self.actorId,
                                         peerid=self.peerid)
        else:
            self.trust = self.handle.get(actorId=self.actorId,
                                         peerid=self.peerid,
                                         token=self.token)
        return self.trust

    def delete(self):
        """ Delete the trust relationship """
        if not self.handle:
            return False
        self.trust = {}
        return self.handle.delete()

    def modify(self, baseuri=None, secret=None, desc=None, approved=None,
               verified=None, verificationToken=None, peer_approved=None):
        if not self.handle:
            logging.debug("Attempted modifcation of trust without handle")
            return False
        if baseuri:
            self.trust["baseuri"] = baseuri
        if secret:
            self.trust["secret"] = secret
        if desc:
            self.trust["desc"] = desc
        if approved is not None:
            self.trust["approved"] = approved
        if verified is not None:
            self.trust["verified"] = verified
        if verificationToken:
            self.trust["verificationToken"] = verificationToken
        if peer_approved is not None:
            self.trust["peer_approved"] = peer_approved
        return self.handle.modify(baseuri=baseuri, secret=secret, desc=desc, approved=approved,
                                  verified=verified, verificationToken=verificationToken,
                                  peer_approved=peer_approved)

    def create(self, baseuri='', type='', relationship='', secret='',
               approved=False, verified=False, verificationToken='',
               desc='', peer_approved=False):
        """ Create a new trust relationship """
        self.trust = {}
        self.trust["baseuri"] = baseuri
        self.trust["type"] = type
        if not relationship or len(relationship) == 0:
            self.trust["relationship"] = self.config.default_relationship
        else:
            self.trust["relationship"] = relationship
        if not secret or len(secret) == 0:
            self.trust["secret"] = self.config.newToken()
        else:
            self.trust["secret"] = secret
        # Be absolutely sure that the secret is not already used
        testhandle = self.config.db_trust.db_trust()
        if testhandle.isTokenInDB(actorId=self.actorId, token=self.trust["secret"]):
            logging.warn("Found a non-unique token where it should be unique")
            return False
        self.trust["approved"] = approved
        self.trust["peer_approved"] = peer_approved
        self.trust["verified"] = verified
        if not verificationToken or len(verificationToken) == 0:
            self.trust["verificationToken"] = self.config.newToken()
        self.trust["desc"] = desc
        self.trust["id"] = self.actorId
        self.trust["peerid"] = self.peerid
        return self.handle.create(actorId=self.actorId,
                                  peerid=self.peerid,
                                  baseuri=self.trust["baseuri"],
                                  type=self.trust["type"],
                                  relationship=self.trust["relationship"],
                                  secret=self.trust["secret"],
                                  approved=self.trust["approved"],
                                  verified=self.trust["verified"],
                                  peer_approved=self.trust["peer_approved"],
                                  verificationToken=self.trust["verificationToken"],
                                  desc=self.trust["desc"])

    def __init__(self, actorId=None, peerid=None, token=None, config=None):
        self.config = config
        self.handle = self.config.db_trust.db_trust()
        self.trust = {}
        if not actorId or len(actorId) == 0:
            logging.debug("No actorid set in initialisation of trust")
            return
        if not peerid and not token:
            logging.debug("Both peerid and token are not set in initialisation of trust. One must be set.")
            return
        if not token and (not peerid or len(peerid) == 0):
            logging.debug("No peerid set in initialisation of trust")
            return
        self.actorId = actorId
        self.peerid = peerid
        self.token = token
        self.get()


class trusts():
    """ Handles all trusts of a specific actor_id

        Access the indvidual trusts in .dbtrusts and the trust data
        in .trusts as a dictionary
    """

    def fetch(self):
        if self.trusts is not None:
            return self.trusts
        if not self.list:
            self.config.db_trust.db_trust_list()
        if not self.trusts:
            self.trusts = self.list.fetch(actorId=self.actorId)
        return self.trusts

    def delete(self):
        if not self.list:
            logging.debug("Already deleted list in trusts")
            return False
        self.list.delete()
        return True

    def __init__(self,  actorId=None, config=None):
        """ Properties must always be initialised with an actorId """
        self.config = config
        if not actorId:
            self.list = None
            logging.debug("No actorId in initialisation of trusts")
            return
        self.list = self.config.db_trust.db_trust_list()
        self.actorId = actorId
        self.trusts = None
        self.fetch()





import logging
import datetime
import os

from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, UTCDateTimeAttribute

"""
    db_subscription_diff handles all db operations for a subscription diff

    db_subscription_diff_list handles list of subscriptions diffs
    Google datastore for google is used as a backend.
"""


class SubscriptionDiff(Model):
    class Meta:
        table_name = "subscriptiondiffs"
        read_capacity_units = 2
        write_capacity_units = 3
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    subid_seqnr = UnicodeAttribute(range_key=True)
    subid = UnicodeAttribute()
    timestamp = UTCDateTimeAttribute(default=datetime.datetime.now())
    diff = UnicodeAttribute()
    seqnr = NumberAttribute(default=1)


class db_subscription_diff():
    """
        db_subscription_diff does all the db operations for subscription diff objects

        The  actorId must always be set.
    """

    def get(self,  actorId=None, subid=None, seqnr=None):
        """ Retrieves the subscriptiondiff from the database """
        if not actorId and not self.handle:
            return None
        if not subid and not self.handle:
            logging.debug("Attempt to get subscriptiondiff without subid")
            return None
        if not self.handle:
            if not seqnr:
                query = SubscriptionDiff.query(
                    actorId,
                    SubscriptionDiff.subid_seqnr.startswith(subid),
                    consistent_read=True)
                # Find the record with lowest seqnr
                for t in query:
                    if not self.handle:
                        self.handle = t
                        continue
                    if t.seqnr < self.handle.seqnr:
                        self.handle = t
            else:
                self.handle = SubscriptionDiff.get(
                    actorId,
                    subid + ":" + unicode(str(seqnr), encoding='UTF-8'),
                    consistent_read=True)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "subscriptionid": t.subid,
                "timestamp": t.timestamp,
                "data": t.diff,
                "sequence": t.seqnr,
            }
        else:
            return None

    def create(self, actorId=None,
               subid=None,
               diff='',
               seqnr=1):
        """ Create a new subscription diff """
        if not actorId or not subid:
            logging.debug("Attempt to create subscriptiondiff without actorid or subid")
            return False
        self.handle = SubscriptionDiff(id=actorId,
                                       subid_seqnr=subid + ":" + unicode(str(seqnr), encoding='UTF-8'),
                                       subid=subid,
                                       diff=diff,
                                       seqnr=seqnr)
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the subscription diff in the database """
        if not self.handle:
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not SubscriptionDiff.exists():
            SubscriptionDiff.create_table(wait=True)


class db_subscription_diff_list():
    """
        db_subscription_diff_list does all the db operations for list of diff objects

        The actorId must always be set. 
    """

    def fetch(self, actorId=None, subid=None):
        """ Retrieves the subscription diffs of an actorId from the database as an array"""
        if not actorId:
            return None
        self.actorId = actorId
        self.subid = subid
        if not subid:
            self.handle = SubscriptionDiff.query(
                actorId,
                consistent_read=True)
        else:
            self.handle = SubscriptionDiff.query(
                actorId,
                SubscriptionDiff.subid.startswith(subid),
                consistent_read=True)
        self.diffs = []
        if self.handle:
            for t in self.handle:
                self.diffs.append(
                {
                    "id": t.id,
                    "subscriptionid": t.subid,
                    "timestamp": t.timestamp,
                    "diff": t.diff,
                    "sequence": t.seqnr,
                })
                sorted(self.diffs, key=lambda diff: diff["sequence"])
            return self.diffs
        else:
            return []

    def delete(self, seqnr=None):
        """ Deletes all the fetched subscription diffs in the database 

            Optional seqnr deletes up to (excluding) a specific seqnr
        """
        if not self.handle:
            return False
        if not seqnr or not isinstance(seqnr, int):
            seqnr = 0
        if not self.subid:
            self.handle = SubscriptionDiff.query(
                self.actorId,
                consistent_read=True)
        else:
            self.handle = SubscriptionDiff.query(
                self.actorId,
                SubscriptionDiff.subid.startswith(self.subid),
                consistent_read=True)
        for p in self.handle:
            if seqnr == 0 or p.seqnr <= seqnr:
                p.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        self.diffs = []
        self.actorId = None
        self.subid = None
        if not SubscriptionDiff.exists():
            SubscriptionDiff.create_table(wait=True)



import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
import logging

"""
    db_actor handles all db operations for an actor
    Google datastore for google is used as a backend.
"""


class CreatorIndex(GlobalSecondaryIndex):
    """
    Secondary index on actor
    """
    class Meta:
        index_name = 'creator-index'
        read_capacity_units = 2
        write_capacity_units = 1
        projection = AllProjection()

    creator = UnicodeAttribute(hash_key=True)


class Actor(Model):
    """
       DynamoDB data model for an actor
    """
    class Meta:
        table_name = "actors"
        read_capacity_units = 6
        write_capacity_units = 2
        region = os.getenv('AWS_DEFAULT_REGION', 'us-west-1')
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    creator = UnicodeAttribute()
    passphrase = UnicodeAttribute()
    creator_index = CreatorIndex()


class db_actor():
    """
        db_actor does all the db operations for actor objects

    """

    def get(self,  actorId=None):
        """ Retrieves the actor from the database """
        if not actorId:
            return None
        try:
            self.handle = Actor.get(actorId, consistent_read=True)
        except Actor.DoesNotExist:
            return None
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "creator": t.creator,
                "passphrase": t.passphrase,
            }
        else:
            return None

    def getByCreator(self, creator=None):
        """ Retrieves the actor from db based on creator field

            Returns None if none was found. If one is found, that one is
            loaded in the object. If more, all are returned.
        """
        if not creator:
            return None
        self.handle = Actor.creator_index.query(creator)
        ret = []
        for c in self.handle:
            logging.warn("    id (" + c.id + ")")
            ret.append(self.get(actorId=c.id))
        return ret

    def modify(self, creator=None, passphrase=None):
        """ Modify an actor """
        if not self.handle:
            logging.debug("Attempted modification of db_actor without db handle")
            return False
        if creator and len(creator) > 0:
            self.handle.creator = creator
        if passphrase and len(passphrase) > 0:
            self.handle.passphrase = passphrase
        self.handle.save()
        return True

    def create(self, actorId=None, creator=None,
               passphrase=None):
        """ Create a new actor """
        if not actorId:
            return False
        if not creator:
            creator = ''
        if not passphrase:
            passphrase = ''
        if self.get(actorId=actorId):
            logging.warn("Trying to create actor that exists(" + actorId + ")")
            return False
        self.handle = Actor(id=actorId,
                            creator=creator,
                            passphrase=passphrase)
        self.handle.save()
        return True

    def delete(self):
        """ Deletes the actor in the database """
        if not self.handle:
            logging.debug("Attempted delete of db_actor without db handle")
            return False
        self.handle.delete()
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
        if not Actor.exists():
            Actor.create_table(wait=True)


class db_actor_list():
    """
        db_actor_list does all the db operations for list of actor objects
    """

    def fetch(self):
        """ Retrieves the actors in the database """
        self.handle = Actor.scan()
        if self.handle:
            ret = []
            for t in self.handle:
                ret.append({
                    "id": t.id,
                    "creator": t.creator,
                    })
            return ret
        else:
            return False

    def __init__(self):
        self.handle = None



import json

from actingweb import auth
from actingweb.handlers import base_handler


class root_handler(base_handler.base_handler):

    def get(self, id):
        if self.request.get('_method') == 'DELETE':
            self.delete(id)
            return
        (myself, check) = auth.init_actingweb(appreq=self, id=id, path='', subpath='',
                                              config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.checkAuthorisation(path='/', method='GET'):
            self.response.set_status(403)
            return
        pair = {
            'id': myself.id,
            'creator': myself.creator,
            'passphrase': myself.passphrase,
        }
        trustee_root = myself.getProperty('trustee_root').value
        if trustee_root and len(trustee_root) > 0:
            pair['trustee_root'] = trustee_root
        out = json.dumps(pair)
        self.response.write(out.encode('utf-8'))
        self.response.headers["Content-Type"] = "application/json"
        self.response.set_status(200)

    def delete(self, id):
        (myself, check) = auth.init_actingweb(appreq=self,
                                              id=id, path='', subpath='', config=self.config)
        if not myself or check.response["code"] != 200:
            return
        if not check.checkAuthorisation(path='/', method='DELETE'):
            self.response.set_status(403)
            return
        self.on_aw.delete_actor()
        myself.delete()
        self.response.set_status(204)
        return




import logging

class peertrustee():

    def get(self):
        if self.peertrustee and len(self.peertrustee) > 0:
            return self.peertrustee
        self.peertrustee = self.handle.get(actorId=self.actorId,
                                           peerid=self.peerid,
                                           type=self.type)
        return self.peertrustee

    def create(self, baseuri=None, passphrase=None):
        if not self.handle:
            self.handle = self.config.db_peertrustee.db_peertrustee()
        if not self.actorId or not self.peerid:
            logging.debug("Attempt to create new peer trustee without actorId or peerid set")
            return False
        if not self.type or len(self.type) == 0:
            logging.debug("Attempt to create peer trustee without type set.")
            return False
        return self.handle.create(actorId=self.actorId,
                                  peerid=self.peerid,
                                  type=self.type,
                                  baseuri=baseuri,
                                  passphrase=passphrase)

    def delete(self):
        if not self.handle:
            logging.debug("Attempt to delete peertrustee without db handle")
            return False
        return self.handle.delete()

    def __init__(self, actorId=None, peerid=None, shorttype=None, type=None, config=None):
        self.config = config
        self.handle = self.config.db_peertrustee.db_peertrustee()
        self.peertrustee = {}
        self.type = None
        if not actorId or len(actorId) == 0:
            logging.debug("No actorid set in initialisation of peertrust")
            return
        if type:
            self.type = type
        elif not type and shorttype:
            if not self.config.actors[shorttype]:
                logging.error('Got request to initialise peer trustee with unknown shorttype(' + shorttype + ')')
                return
            self.type = self.config.actors[shorttype]["type"]
        elif not peerid:
            logging.debug("Peerid and shorttype are not set in initialisation of peertrustee. One is required")
            return
        self.actorId = actorId
        self.peerid = peerid
        self.get()



import os
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, JSONAttribute, UTCDateTimeAttribute

"""
    db_attribute handles all db operations for an attribute (internal)
    AWS DynamoDB is used as a backend.
"""

class Attribute(Model):
    """
       DynamoDB data model for a property
    """
    class Meta:
        table_name = "attributes"
        read_capacity_units = 26
        write_capacity_units = 2
        host = os.getenv('AWS_DB_HOST', None)

    id = UnicodeAttribute(hash_key=True)
    bucket_name = UnicodeAttribute(range_key=True)
    bucket = UnicodeAttribute()
    name = UnicodeAttribute()
    data = JSONAttribute(null=True)
    timestamp = UTCDateTimeAttribute(null=True)


class db_attribute():
    """
        db_property does all the db operations for property objects

        The actorId must always be set. get(), set() will set a new internal handle
        that will be reused by set() (overwrite attribute) and
        delete().
    """

    def get_bucket(self,  actorId=None, bucket=None):
        """ Returns a dict of attributes from a bucket, each with data and timestamp """
        if not actorId or not bucket:
            return None
        try:
            query = Attribute.query(
                actorId,
                Attribute.bucket_name.startswith(bucket),
                consistent_read=True)
        except Attribute.DoesNotExist:
            return None
        ret = {}
        for t in query:
            ret[t.name] = {
                "data": t.data,
                "timestamp": t.timestamp,
            }
        return ret

    def get_attr(self,  actorId=None, bucket=None, name=None):
        """ Returns a dict of attributes from a bucket, each with data and timestamp """
        if not actorId or not bucket or not name:
            return None
        try:
            r = Attribute.get(actorId, bucket + ":" + name, consistent_read=True)
        except Attribute.DoesNotExist:
            return None
        return {
            "data": r.data,
            "timestamp": r.timestamp,
        }

    def set_attr(self, actorId=None, bucket=None, name=None, data=None, timestamp=None):
        """ Sets a data value for a given attribute in a bucket
        """
        if not actorId or not name or not bucket:
            return False
        if not data or len(data) == 0:
            try:
                item = Attribute.get(actorId, bucket + ":" + name, consistent_read=True)
                item.delete()
            except Attribute.DoesNotExist:
                pass
            return True
        new = Attribute(
            id=actorId,
            bucket_name=bucket + ":" + name,
            bucket=bucket,
            name=name,
            data=data,
            timestamp=timestamp
        )
        new.save()
        return True

    def delete_attr(self, actorId=None, bucket=None, name=None):
        """ Deletes an attribute in a bucket
        """
        return self.set_attr(actorId=actorId, bucket=bucket, name=name, data=None)

    def delete_bucket(self, actorId=None, bucket=None):
        """ Deletes an entire bucket
        """
        if not actorId or not bucket:
            return False
        try:
            query = Attribute.query(
                actorId,
                Attribute.bucket_name.startswith(bucket),
                consistent_read=True)
        except Attribute.DoesNotExist:
            return True
        for t in query:
            t.delete()
        return True

    def __init__(self):
        if not Attribute.exists():
            Attribute.create_table(wait=True)


class db_attribute_bucket_list():
    """
        db_attribute_bucket_list handles multiple buckets

        The  actorId must always be set.
    """

    def fetch(self,  actorId=None):
        """ Retrieves all the attributes of an actorId from the database """
        if not actorId:
            return None
        try:
            query = Attribute.query(actorId)
        except Attribute.DoesNotExist:
            return None
        ret = {}
        for t in query:
            if t.bucket not in ret:
                ret[t.bucket] = {}
            ret[t.bucket][t.name] = {
                "data": t.data,
                "timestamp": t.timestamp,
            }
        return ret

    def delete(self, actorId=None):
        """ Deletes all the attributes in the database """
        if not actorId:
            return False
        try:
            query = Attribute.query(actorId)
        except Attribute.DoesNotExist:
            return False
        for t in query:
            t.delete()
        return True

    def __init__(self):
        if not Attribute.exists():
            Attribute.create_table(wait=True)


class property():
    """
        property is the main entity keeping a property.

        It needs to be initalised at object creation time.

    """

    def get(self):
        """ Retrieves the property from the database """
        if not self.dbprop:
            # New property after a delete()
            self.dbprop = self.config.db_property.db_property()
            self.value=None
        self.value = self.dbprop.get(actorId=self.actorId, name=self.name)
        return self.value

    def set(self, value):
        """ Sets a new value for this property """
        if not self.dbprop:
            # New property after a delete()
            self.dbprop = self.config.db_property.db_property()
        if not self.actorId or not self.name:
            return False
        # Make sure we have made a dip in db to avoid two properties
        # with same name
        db_value = self.dbprop.get(actorId=self.actorId, name=self.name)
        if db_value == value:
            return True
        self.value = value
        return self.dbprop.set(actorId=self.actorId, name=self.name, value=value)

    def delete(self):
        """ Deletes the property in the database """
        if not self.dbprop:
            return
        if self.dbprop.delete():
            self.value = None
            self.dbprop = None
            return True
        else:
            return False

    def getActorId(self):
        return self.actorId

    def __init__(self,  actorId=None, name=None, value=None, config=None):
        """ A property must be initialised with actorId and name or
            name and value (to find an actor's property of a certain value)
        """
        self.config = config
        self.dbprop = self.config.db_property.db_property()
        self.name = name
        if not actorId and name and len(name) > 0 and value and len(value) > 0:
            self.actorId = self.dbprop.get_actorId_from_property(name=name,
                                                                 value=value)
            if not self.actorId:
                return
            self.value = value
        else:
            self.actorId = actorId
            self.value = None
            if name and len(name) > 0:
                self.get()


class properties():
    """ Handles all properties of a specific actor_id

        Access the properties
        in .props as a dictionary
    """

    def fetch(self):
        if not self.actorId:
            return False
        if not self.list:
            return False
        if self.props is not None:
            return self.props
        self.props = self.list.fetch(actorId=self.actorId)
        return self.props

    def delete(self):
        if not self.list:
            self.fetch()
        if not self.list:
            return False
        self.list.delete()
        return True

    def __init__(self,  actorId=None, config=None):
        """ Properties must always be initialised with an actorId """
        self.config = config
        if not actorId:
            self.list = None
            return False
        self.list = self.config.db_property.db_property_list()
        self.actorId = actorId
        self.props = None
        self.fetch()





from google.appengine.ext import ndb
import logging

"""
    db_subscription handles all db operations for a subscription

    db_subscription_list handles list of subscriptions
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_subscription',
    'db_subscription_list',
]


class Subscription(ndb.Model):
    id = ndb.StringProperty(required=True)
    peerid = ndb.StringProperty(required=True)
    subid = ndb.StringProperty(required=True)
    granularity = ndb.StringProperty()
    target = ndb.StringProperty()
    subtarget = ndb.StringProperty()
    resource = ndb.StringProperty()
    seqnr = ndb.IntegerProperty(default=1)
    callback = ndb.BooleanProperty()


class db_subscription():
    """
        db_subscription does all the db operations for subscription objects

        The  actorId must always be set.
    """

    def get(self,  actorId=None, peerid=None, subid=None):
        """ Retrieves the subscription from the database """
        if not actorId:
            return None
        if not peerid or not subid:
            logging.debug("Attempt to get subscription without peerid or subid")
            return None
        if not self.handle:
            self.handle = Subscription.query(Subscription.id == actorId,
                                             Subscription.peerid == peerid,
                                             Subscription.subid == subid).get(use_cache=False)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "peerid": t.peerid,
                "subscriptionid": t.subid,
                "granularity": t.granularity,
                "target": t.target,
                "subtarget": t.subtarget,
                "resource": t.resource,
                "sequence": t.seqnr,
                "callback": t.callback,
            }
        else:
            return None

    def modify(self, actorId=None,
               peerid=None,
               subid=None,
               granularity=None,
               target=None,
               subtarget=None,
               resource=None,
               seqnr=None,
               callback=None):
        """ Modify a subscription

            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of db_subscription without db handle")
            return False
        if peerid and len(peerid) > 0:
            self.handle.peerid = peerid
        if subid and len(subid) > 0:
            self.handle.subid = subid
        if granularity and len(granularity) > 0:
            self.handle.granularity = granularity
        if callback is not None:
            self.handle.callback = callback
        if target and len(target) > 0:
            self.handle.target = target
        if subtarget and len(subtarget) > 0:
            self.handle.subtarget = subtarget
        if resource and len(resource) > 0:
            self.handle.resource = resource
        if seqnr:
            self.handle.seqnr = seqnr
        self.handle.put(use_cache=False)
        return True

    def create(self, actorId=None,
               peerid=None,
               subid=None,
               granularity=None,
               target=None,
               subtarget=None,
               resource=None,
               seqnr=None,
               callback=None):
        """ Create a new subscription """
        if not actorId or not peerid or not subid:
            return False
        if not granularity:
            granularity = ''
        if not target:
            target = ''
        if not subtarget:
            subtarget = ''
        if not resource:
            resource = ''
        if not seqnr:
            seqnr = 1
        if not callback:
            callback = False
        self.handle = Subscription(id=actorId,
                                   peerid=peerid,
                                   subid=subid,
                                   granularity=granularity,
                                   target=target,
                                   subtarget=subtarget,
                                   resource=resource,
                                   seqnr=seqnr,
                                   callback=callback)
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the subscription in the database """
        if not self.handle:
            logging.debug("Attempted delete of db_subscription with no handle set.")
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None


class db_subscription_list():
    """
        db_trust_list does all the db operations for list of trust objects

        The  actorId must always be set.
    """

    def fetch(self, actorId):
        """ Retrieves the subscriptions of an actorId from the database as an array"""
        if not actorId:
            return None
        self.handle = Subscription.query(Subscription.id == actorId).fetch(use_cache=False)
        self.subscriptions = []
        if self.handle:
            for t in self.handle:
                self.subscriptions.append(
                {
                    "id": t.id,
                    "peerid": t.peerid,
                    "subscriptionid": t.subid,
                    "granularity": t.granularity,
                    "target": t.target,
                    "subtarget": t.subtarget,
                    "resource": t.resource,
                    "sequence": t.seqnr,
                    "callback": t.callback,
                })
            return self.subscriptions
        else:
            return []

    def delete(self):
        """ Deletes all the subscriptions for an actor in the database """
        if not self.handle:
            return False
        for p in self.handle:
            p.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None



from actingweb import aw_web_request, on_aw, config

class base_handler():

    def __init__(self,
                 webobj=aw_web_request.aw_webobj(),
                 config=config.config(),
                 on_aw=on_aw.on_aw_base()
                 ):
        self.request = webobj.request
        self.response = webobj.response
        self.config = config
        self.on_aw = on_aw




from google.appengine.ext import ndb
import logging

"""
    db_subscription_diff handles all db operations for a subscription diff

    db_subscription_diff_list handles list of subscriptions diffs
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_subscription_diff',
    'db_subscription_diff_list',
]


class SubscriptionDiff(ndb.Model):
    id = ndb.StringProperty(required=True)
    subid = ndb.StringProperty(required=True)
    timestamp = ndb.DateTimeProperty(auto_now_add=True)
    diff = ndb.TextProperty()
    seqnr = ndb.IntegerProperty()


class db_subscription_diff():
    """
        db_subscription_diff does all the db operations for subscription diff objects

        The  actorId must always be set.
    """

    def get(self,  actorId=None, subid=None, seqnr=None):
        """ Retrieves the subscriptiondiff from the database """
        if not actorId and not self.handle:
            return None
        if not subid and not self.handle:
            logging.debug("Attempt to get subscriptiondiff without subid")
            return None
        if not self.handle:
            if not seqnr:
                self.handle = SubscriptionDiff.query(SubscriptionDiff.id == actorId,
                                                     SubscriptionDiff.subid == subid
                                                     ).get(use_cache=False)
            else:
                self.handle = SubscriptionDiff.query(SubscriptionDiff.id == actorId,
                                                     SubscriptionDiff.subid == subid,
                                                     SubscriptionDiff.seqnr == seqnr
                                                     ).get(use_cache=False)
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "subscriptionid": t.subid,
                "timestamp": t.timestamp,
                "data": t.diff,
                "sequence": t.seqnr,
            }
        else:
            return None

    def create(self, actorId=None,
               subid=None,
               diff=None,
               seqnr=None):
        """ Create a new subscription diff """
        if not actorId or not subid:
            logging.debug("Attempt to create subscriptiondiff without actorid or subid")
            return False
        if not seqnr:
            seqnr = 1
        if not diff:
            diff = ''
        self.handle = SubscriptionDiff(id=actorId,
                                       subid=subid,
                                       diff=diff,
                                       seqnr=seqnr)
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the subscription diff in the database """
        if not self.handle:
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None


class db_subscription_diff_list():
    """
        db_subscription_diff_list does all the db operations for list of diff objects

        The actorId must always be set. 
    """

    def fetch(self, actorId=None, subid=None):
        """ Retrieves the subscription diffs of an actorId from the database as an array"""
        if not actorId:
            return None
        if not subid:
            self.handle = SubscriptionDiff.query(SubscriptionDiff.id == actorId).order(SubscriptionDiff.seqnr).fetch(use_cache=False)
        else:
            self.handle = SubscriptionDiff.query(SubscriptionDiff.id == actorId,
                                                 SubscriptionDiff.subid == subid).order(SubscriptionDiff.seqnr).fetch(use_cache=False)
        self.diffs = []
        if self.handle:
            for t in self.handle:
                self.diffs.append(
                {
                    "id": t.id,
                    "subscriptionid": t.subid,
                    "timestamp": t.timestamp,
                    "diff": t.diff,
                    "sequence": t.seqnr,
                })
            return self.diffs
        else:
            return []

    def delete(self, seqnr=None):
        """ Deletes all the fetched subscription diffs in the database 

            Optional seqnr deletes up to (excluding) a specific seqnr
        """
        if not self.handle:
            return False
        if not seqnr or not isinstance(seqnr, int):
            seqnr = 0
        for p in self.handle:
            if seqnr == 0 or p.seqnr <= seqnr:
                p.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None



from google.appengine.ext import ndb

"""
    db_peertrustee handles all db operations for a peer we are a trustee for.
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_peertrustee',
    'db_peertrustee_list',
]


class PeerTrustee(ndb.Model):
    id = ndb.StringProperty(required=True)
    peerid = ndb.StringProperty(required=True)
    baseuri = ndb.StringProperty(required=True)
    type = ndb.StringProperty(required=True)
    passphrase = ndb.StringProperty(required=True)


class db_peertrustee():
    """
        db_peertrustee does all the db operations for property objects

        The actorId must always be set.
    """

    def get(self,  actorId=None, type=None, peerid=None):
        """ Retrieves the peertrustee from the database """
        if not actorId:
            return None
        if not peerid and not type:
            logging.debug("Attempt to get db_peertrustee without peerid or type")
            return None
        if not self.handle and peerid:
            self.handle = PeerTrustee.query(PeerTrustee.id == actorId,
                                            PeerTrustee.peerid == peerid).get(use_cache=False)
        elif not self.handle and type:
            self.handle = PeerTrustee.query(PeerTrustee.id == actorId,
                                            PeerTrustee.type == type).fetch(use_cache=False)
            if len(self.handle) > 1:
                logging.error('Found more than one peer of this peer trustee type(' + 
                              shorttype + '). Unable to determine which, need peerid lookup.')
                return False
            if len(self.handle) == 1:
                self.handle = self.handle[0]
        if self.handle:
            t = self.handle
            return {
                "id": t.id,
                "peerid": t.peerid,
                "baseuri": t.baseuri,
                "type": t.type,
                "passphrase": t.passphrase,
            }
        else:
            return None

    def create(self, actorId=None,
               peerid=None,
               type=None,
               baseuri=None,
               passphrase=None):
        """ Create a new peertrustee """
        if not actorId or not peerid or not type:
            logging.debug("actorId, peerid, and type are mandatory when creating peertrustee in db")
            return False
        if not baseuri:
            baseuri = ''
        if not passphrase:
            passphrase = ''
        self.handle = PeerTrustee(id=actorId,
                                  peerid=peerid,
                                  type=type,
                                  baseuri=baseuri,
                                  passphrase=passphrase)
        self.handle.put(use_cache=False)
        return True

    def modify(self,
               type=None,
               baseuri=None,
               passphrase=None):
        """ Modify a peertrustee

            If bools are none, they will not be changed.
        """
        if not self.handle:
            logging.debug("Attempted modification of db_peertrustee without db handle")
            return False
        if baseuri and len(baseuri) > 0:
            self.handle.baseuri = baseuri
        if passphrase and len(passphrase) > 0:
            self.handle.passphrase = passphrase
        if type and len(type) > 0:
            self.handle.type = type
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the peertrustee in the database after a get() """
        if not self.handle:
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None


class db_peertrustee_list():
    """
        db_peertrustee_list does all the db operations for list of peertrustee objects

        The  actorId must always be set.
    """

    def fetch(self,  actorId=None):
        """ Retrieves the peer trustees of an actorId from the database """
        if not actorId:
            return None
        self.handle = PeerTrustee.query(PeerTrustee.id == actorId).fetch(use_cache=False)
        self.peertrustees = []
        if self.handle:
            for t in self.handle:
                self.peertrustees.append(
                {
                  "id": t.id,
                  "peerid": t.peerid,
                  "baseuri": t.baseuri,
                  "type": t.type,
                  "passphrase": t.passphrase,
                })
            return self.peertrustees
        else:
            return []

    def delete(self):
        """ Deletes all the peertrustees in the database """
        if not self.handle:
            return False
        for p in self.handle:
            p.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None



from google.appengine.ext import ndb

"""
    db_property handles all db operations for a property
    Google datastore for google is used as a backend.
"""

__all__ = [
    'db_property',
    'db_property_list',
]


class Property(ndb.Model):
    """
       NDB data model for a property
    """
    id = ndb.StringProperty(required=True)
    name = ndb.StringProperty(required=True)
    value = ndb.TextProperty()


class db_property():
    """
        db_property does all the db operations for property objects

        The actorId must always be set. get(), set() and
        get_actorId_from_property() will set a new internal handle
        that will be reused by set() (overwrite property) and
        delete().
    """

    def get(self,  actorId=None, name=None):
        """ Retrieves the property from the database """
        if not actorId or not name:
            return None
        if self.handle and self.handle.value:
            return self.handle.value
        self.handle = Property.query(Property.id == actorId,
                                     Property.name == name).get(use_cache=False)
        if self.handle:
            return self.handle.value
        else:
            return None

    def get_actorId_from_property(self, name=None, value=None):
        """ Retrives an actorId based on the value of a property.

        Note that this is a costly operation as all properties of this type
        must be retrieved and processed as value is TextProperty and cannot
        be indexed.
        """
        if not name or not value:
            return None
        results = Property.query(Property.name == name).fetch(use_cache=False)
        self.handle = None
        for res in results:
            if res.value != value:
                continue
            self.handle = res
            break
        if not self.handle:
            return None
        return self.handle.id

    def set(self, actorId=None, name=None, value=None):
        """ Sets a new value for the property name

            If get() has not been run first, a new property will be created
            with the same name.
        """
        if not name:
            return False
        if not value:
            value = ''
        if not self.handle:
            if not actorId:
                return False
            self.handle = Property(id=actorId, name=name,
                                   value=value)
        else:
            self.handle.value = value
        self.handle.put(use_cache=False)
        return True

    def delete(self):
        """ Deletes the property in the database after a get() """
        if not self.handle:
            return False
        self.handle.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None


class db_property_list():
    """
        db_property does all the db operations for list of property objects

        The  actorId must always be set.
    """

    def fetch(self,  actorId=None):
        """ Retrieves the properties of an actorId from the database """
        if not actorId:
            return None
        self.handle = Property.query(Property.id == actorId).fetch(use_cache=False)
        if self.handle:
            self.props = {}
            for d in self.handle:
                self.props[d.name] = d.value
            return self.props
        else:
            return None

    def delete(self):
        """ Deletes all the properties in the database """
        if not self.handle:
            return False
        for p in self.handle:
            p.key.delete(use_cache=False)
        self.handle = None
        return True

    def __init__(self):
        self.handle = None
